@@ -25,10 +25,11 @@ pub(crate) struct Config {
   pub(crate) search_config:        SearchConfig,
   pub(crate) shell:                String,
   pub(crate) shell_args:           Vec<String>,
-  pub(crate) shell_present:        bool,
   pub(crate) shell_command:        bool,
+  pub(crate) shell_present:        bool,
   pub(crate) subcommand:           Subcommand,
   pub(crate) unsorted:             bool,
+  pub(crate) unstable:             bool,
   pub(crate) verbosity:            Verbosity,
 }
 
@@ -38,6 +39,7 @@ mod cmd {
   pub(crate) const DUMP: &str = "DUMP";
   pub(crate) const EDIT: &str = "EDIT";
   pub(crate) const EVALUATE: &str = "EVALUATE";
+  pub(crate) const FORMAT: &str = "FORMAT";
   pub(crate) const INIT: &str = "INIT";
   pub(crate) const LIST: &str = "LIST";
   pub(crate) const SHOW: &str = "SHOW";
@@ -52,6 +54,7 @@ mod cmd {
     DUMP,
     EDIT,
     EVALUATE,
+    FORMAT,
     INIT,
     LIST,
     SHOW,
@@ -63,6 +66,7 @@ mod cmd {
     COMPLETIONS,
     DUMP,
     EDIT,
+    FORMAT,
     INIT,
     LIST,
     SHOW,
@@ -89,6 +93,7 @@ mod arg {
   pub(crate) const SHELL_ARG: &str = "SHELL-ARG";
   pub(crate) const SHELL_COMMAND: &str = "SHELL-COMMAND";
   pub(crate) const UNSORTED: &str = "UNSORTED";
+  pub(crate) const UNSTABLE: &str = "UNSTABLE";
   pub(crate) const VERBOSE: &str = "VERBOSE";
   pub(crate) const WORKING_DIRECTORY: &str = "WORKING-DIRECTORY";
 
@@ -215,6 +220,11 @@ impl Config {
           .short("u")
           .help("Return list and summary entries in source order"),
       )
+      .arg(
+        Arg::with_name(arg::UNSTABLE)
+          .long("unstable")
+          .help("Enable unstable features"),
+      )
       .arg(
         Arg::with_name(arg::VERBOSE)
           .short("v")
@@ -266,6 +276,11 @@ impl Config {
         "Evaluate and print all variables. If a variable name is given as an argument, only print \
          that variable's value.",
       ))
+      .arg(
+        Arg::with_name(cmd::FORMAT)
+          .long("fmt")
+          .help("Format and overwrite justfile"),
+      )
       .arg(
         Arg::with_name(cmd::INIT)
           .long("init")
@@ -310,7 +325,7 @@ impl Config {
       ))
     } else {
       app
-        .version(concat!("v", env!("CARGO_PKG_VERSION")))
+        .version(env!("CARGO_PKG_VERSION"))
         .author(env!("CARGO_PKG_AUTHORS"))
         .about(concat!(
           env!("CARGO_PKG_DESCRIPTION"),
@@ -442,6 +457,8 @@ impl Config {
       Subcommand::Summary
     } else if matches.is_present(cmd::DUMP) {
       Subcommand::Dump
+    } else if matches.is_present(cmd::FORMAT) {
+      Subcommand::Format
     } else if matches.is_present(cmd::INIT) {
       Subcommand::Init
     } else if matches.is_present(cmd::LIST) {
@@ -496,6 +513,7 @@ impl Config {
       load_dotenv: !matches.is_present(arg::NO_DOTENV),
       shell_command: matches.is_present(arg::SHELL_COMMAND),
       unsorted: matches.is_present(arg::UNSORTED),
+      unstable: matches.is_present(arg::UNSTABLE),
       list_heading: matches
         .value_of(arg::LIST_HEADING)
         .unwrap_or("Available recipes:\n")
@@ -539,7 +557,9 @@ impl Config {
       })
       .eprint(self.color)?;
 
-    let justfile = Compiler::compile(&src).eprint(self.color)?;
+    let tokens = Lexer::lex(&src).eprint(self.color)?;
+    let ast = Parser::parse(&tokens).eprint(self.color)?;
+    let justfile = Analyzer::analyze(ast.clone()).eprint(self.color)?;
 
     if self.verbosity.loud() {
       for warning in &justfile.warnings {
@@ -555,8 +575,9 @@ impl Config {
       Choose { overrides, chooser } =>
         self.choose(justfile, &search, overrides, chooser.as_deref())?,
       Command { overrides, .. } => self.run(justfile, &search, overrides, &[])?,
-      Dump => Self::dump(justfile),
+      Dump => Self::dump(ast)?,
       Evaluate { overrides, .. } => self.run(justfile, &search, overrides, &[])?,
+      Format => self.format(ast, &search)?,
       List => self.list(justfile),
       Run {
         arguments,
@@ -676,8 +697,9 @@ impl Config {
     self.run(justfile, search, overrides, &recipes)
   }
 
-  fn dump(justfile: Justfile) {
-    println!("{}", justfile);
+  fn dump(ast: Module) -> Result<(), i32> {
+    print!("{}", ast);
+    Ok(())
   }
 
   pub(crate) fn edit(&self, search: &Search) -> Result<(), i32> {
@@ -713,6 +735,32 @@ impl Config {
     }
   }
 
+  fn format(&self, ast: Module, search: &Search) -> Result<(), i32> {
+    if !self.unstable {
+      eprintln!(
+        "The `--fmt` command is currently unstable. Pass the `--unstable` flag to enable it."
+      );
+      return Err(EXIT_FAILURE);
+    }
+
+    if let Err(error) = File::create(&search.justfile).and_then(|mut file| write!(file, "{}", ast))
+    {
+      if self.verbosity.loud() {
+        eprintln!(
+          "Failed to write justfile to `{}`: {}",
+          search.justfile.display(),
+          error
+        );
+      }
+      Err(EXIT_FAILURE)
+    } else {
+      if self.verbosity.loud() {
+        eprintln!("Wrote justfile to `{}`", search.justfile.display());
+      }
+      Ok(())
+    }
+  }
+
   pub(crate) fn init(&self) -> Result<(), i32> {
     let search =
       Search::init(&self.search_config, &self.invocation_directory).eprint(self.color)?;
@@ -900,7 +948,7 @@ mod tests {
   // have proper tests for all the flags, but this will do for now.
   #[test]
   fn help() {
-    const EXPECTED_HELP: &str = "just v0.9.4
+    const EXPECTED_HELP: &str = "just 0.9.5
 Casey Rodarmor <casey@rodarmor.com>
 🤖 Just a command runner \
                                  - https://github.com/casey/just
@@ -920,6 +968,7 @@ FLAGS:
         --evaluate            Evaluate and print all variables. If a variable name is given as an \
                                  argument, only print
                               that variable's value.
+        --fmt                 Format and overwrite justfile
         --highlight           Highlight echoed recipe lines in bold
         --init                Initialize new justfile in project root
     -l, --list                List available recipes and their arguments
@@ -930,6 +979,7 @@ FLAGS:
                                  backticks
         --summary             List names of available recipes
     -u, --unsorted            Return list and summary entries in source order
+        --unstable            Enable unstable features
         --variables           List names of variables
     -v, --verbose             Use verbose output
 
@@ -1315,6 +1365,11 @@ ARGS:
     args: ["--list", "--dump"],
   }
 
+  error! {
+    name: subcommand_conflict_fmt,
+    args: ["--list", "--fmt"],
+  }
+
   error! {
     name: subcommand_conflict_init,
     args: ["--list", "--init"],
@@ -1661,6 +1716,16 @@ ARGS:
     },
   }
 
+  error! {
+    name: fmt_arguments,
+    args: ["--fmt", "bar"],
+    error: ConfigError::SubcommandArguments { subcommand, arguments },
+    check: {
+      assert_eq!(subcommand, cmd::FORMAT);
+      assert_eq!(arguments, &["bar"]);
+    },
+  }
+
   error! {
     name: init_arguments,
     args: ["--init", "bar"],
